import { ReusablePipeline } from './ReusablePipeline';
import { IntermediateStage, TerminalStage } from './stages';
/**
 * A reusable pipeline that lazily evaluates the data processing operations which are referred to as stages.
 * These stages are only evaluated when a terminal stage is triggered by calling
 * {@link collect()}, passing in an implementation of a terminal operation or by calling or {@link toArray()}.
 *
 * Each lazy pipeline is a stage itself that acts as the source which in turn pipes
 * each element in the underlying data source to its downstream stages.
 *
 * A notable feature of this pipeline is that it is reusable, so you can construct your pipeline however
 * you'd like and reuse it in different places with no worries about data corruptions by calling {@link resume()}
 * before triggering a terminal stage on the pipeline, otherwise, an error will occur.
 *
 * @template IN The type parameter of each element in the pipeline.
 */
export declare class LazyPipeline<IN> extends IntermediateStage<IN, any> implements ReusablePipeline<IN> {
    /**
     * The intermediate stages that were added to this pipeline so far
     */
    private readonly _stages;
    /**
     * The data source to operate upon
     */
    private _source;
    private _isTerminated;
    private _isFrozen;
    private _terminalStage;
    constructor(source: Iterable<IN>);
    static from<IN>(iterable: Iterable<IN>): LazyPipeline<IN>;
    add<A>(stage: IntermediateStage<IN, A>): ReusablePipeline<A>;
    add<A, B>(stage1: IntermediateStage<IN, A>, stage2: IntermediateStage<A, B>): ReusablePipeline<B>;
    add<A, B, C>(stage1: IntermediateStage<IN, A>, stage2: IntermediateStage<A, B>, stage3: IntermediateStage<B, C>): ReusablePipeline<C>;
    add<A, B, C, D>(stage1: IntermediateStage<IN, A>, stage2: IntermediateStage<A, B>, stage3: IntermediateStage<B, C>, stage4: IntermediateStage<C, D>): ReusablePipeline<D>;
    add<A, B, C, D, E>(stage1: IntermediateStage<IN, A>, stage2: IntermediateStage<A, B>, stage3: IntermediateStage<B, C>, stage4: IntermediateStage<C, D>, stage5: IntermediateStage<D, E>): ReusablePipeline<E>;
    add<A, B, C, D, E, F>(stage1: IntermediateStage<IN, A>, stage2: IntermediateStage<A, B>, stage3: IntermediateStage<B, C>, stage4: IntermediateStage<C, D>, stage5: IntermediateStage<D, E>, stage6: IntermediateStage<E, F>): ReusablePipeline<F>;
    add<A, B, C, D, E, F, G>(stage1: IntermediateStage<IN, A>, stage2: IntermediateStage<A, B>, stage3: IntermediateStage<B, C>, stage4: IntermediateStage<C, D>, stage5: IntermediateStage<D, E>, stage6: IntermediateStage<E, F>, stage7: IntermediateStage<F, G>): ReusablePipeline<G>;
    add<A, B, C, D, E, F, G, H>(stage1: IntermediateStage<IN, A>, stage2: IntermediateStage<A, B>, stage3: IntermediateStage<B, C>, stage4: IntermediateStage<C, D>, stage5: IntermediateStage<D, E>, stage6: IntermediateStage<E, F>, stage7: IntermediateStage<F, G>, stage8: IntermediateStage<G, H>): ReusablePipeline<H>;
    add<A, B, C, D, E, F, G, H, I>(stage1: IntermediateStage<IN, A>, stage2: IntermediateStage<A, B>, stage3: IntermediateStage<B, C>, stage4: IntermediateStage<C, D>, stage5: IntermediateStage<D, E>, stage6: IntermediateStage<E, F>, stage7: IntermediateStage<F, G>, stage8: IntermediateStage<G, H>, stage9: IntermediateStage<H, I>): ReusablePipeline<I>;
    add<A, B, C, D, E, F, G, H, I, J>(stage1: IntermediateStage<IN, A>, stage2: IntermediateStage<A, B>, stage3: IntermediateStage<B, C>, stage4: IntermediateStage<C, D>, stage5: IntermediateStage<D, E>, stage6: IntermediateStage<E, F>, stage7: IntermediateStage<F, G>, stage8: IntermediateStage<G, H>, stage9: IntermediateStage<H, I>, stage10: IntermediateStage<I, J>, ...stages: IntermediateStage<J, any>[]): ReusablePipeline<any>;
    private _attachNewStages;
    private _watchForPipelineEvent;
    collect<OUT>(terminalStage: TerminalStage<IN, OUT>): OUT;
    toArray(): IN[];
    forEach(action: (element: IN) => void): undefined;
    resume(): void;
    private _reconnectAllStages;
    readFrom(newSource: Iterable<IN>): void;
    freeze(): void;
    unfreeze(): void;
    consume(element: IN, hasMoreDataUpstream: boolean): void;
}
